# Charge stability diagram for zoom and large scan

# %% Import Declaration
import matplotlib.pyplot as plt
import setuptools
import seaborn as sns
import matplotlib as mpl
import os
import numpy as np
import pandas as pd
import scipy as sp
import skrf as rf
from qutip import *
from numpy import pi

from scipy.optimize import curve_fit
from colour import Color
from matplotlib.colors import LinearSegmentedColormap

from lmfit.models import LorentzianModel, GaussianModel, ConstantModel, LinearModel
from lmfit import fit_report

from matplotlib.ticker import AutoMinorLocator


# %% Global plt.plot Setting 

def rotate_matrix (x, y, angle, x_shift=0, y_shift=0, units="DEGREES"):
    """
    Rotates a point in the xy-plane counterclockwise through an angle about the origin
    https://en.wikipedia.org/wiki/Rotation_matrix
    :param x: x coordinate
    :param y: y coordinate
    :param x_shift: x-axis shift from origin (0, 0)
    :param y_shift: y-axis shift from origin (0, 0)
    :param angle: The rotation angle in degrees
    :param units: DEGREES (default) or RADIANS
    :return: Tuple of rotated x and y
    """

    # Shift to origin (0,0)
    x = x - x_shift
    y = y - y_shift

    # Convert degrees to radians
    if units == "DEGREES":
        angle = np.radians(angle)

    # Rotation matrix multiplication to get rotated x & y
    xr = (x * np.cos(angle)) - (y * np.sin(angle)) + x_shift
    yr = (x * np.sin(angle)) + (y * np.cos(angle)) + y_shift

    return xr, yr


def make_freq_shift_plot(dataset, Phase_plot=False, vline=0, hline=0):
    filename = str(dataset) + "/frequency_RP2.dat"
    data_df = pd.read_csv(filename, skiprows=[1,2], sep='\t')

    extent = [data_df['# frequency'].min(), data_df['# frequency'].max(), data_df['RP2'].min(), data_df['RP2'].max()]
    cmap = mpl.cm.viridis

    if Phase_plot == True:
        I = data_df['I'].to_numpy()
        Q = data_df['Q'].to_numpy()
        s21_array = np.rad2deg(np.arctan(np.divide(I,Q)))
    elif Phase_plot == False:
        s21_array = data_df['S21mag'].to_numpy()
    
    
    S21_reshape = np.transpose(np.reshape(s21_array,(100,100))) # Reshape array for Large array
    # S21_reshape = (np.reshape(s21_array,(100,100))) # Reshape array for Large array


    fig = plt.figure(figsize=(4.5, 6))
    plt.imshow(S21_reshape, origin='upper', extent=extent, aspect = 'auto', cmap=cmap)
    plt.colorbar(label="", orientation="vertical")
    
    plt.axvline(x=vline, color='r', linestyle='--')
    plt.axhline(y=hline, color='k', linestyle='--')
    
    plt.xlabel('$freq$ [Hz]')
    plt.ylabel('$V_{RP}$ [mV]')
    
    return fig

def make_freq_shift_plot_df(data_df, Phase_plot=False, vline=0, hline=0):

    extent = [data_df[:,0].min(), data_df[:,0].max(), data_df[:,1].min(), data_df[:,1].max()]
    cmap = mpl.cm.viridis

    if Phase_plot == True:
        I = data_df['I'].to_numpy()
        Q = data_df['Q'].to_numpy()
        s21_array = np.rad2deg(np.arctan(np.divide(I,Q)))
    elif Phase_plot == False:
        s21_array = data_df[:,4]
    
    # Grx_to_vdev = 10**((107-17-63)/20) 
    Grx_to_vdev = 1
    S21_reshape = np.transpose(np.reshape(s21_array,(100,100))) * Grx_to_vdev


    fig = plt.figure(figsize=(4.5, 6))
    plt.imshow(S21_reshape, origin='lower', extent=extent, aspect = 'auto', cmap=cmap)
    plt.colorbar(label="", orientation="vertical")
    
    plt.axvline(x=vline, color='r', linestyle='--')
    plt.axhline(y=hline, color='#FFFFFF', linestyle='--')
    
    plt.xlabel('$Freq.$ [Hz]')
    plt.ylabel('$V_{RP}$ [mV]')
    
    plt.xlim([6.89E9,6.92E9])
    return fig


def fit_doubletanh(xdata, ydata, handle_bad_fit=False):
    """ Fits the data using a symmetric double hyperbolic tangent function to 
    find the zero-detuning point of a dispersive shift vs detuning dataset.

    """
    # Normalization and initial values
    x_scale = 10**np.floor(np.log10(np.max(np.abs(xdata))))
    xdata = xdata/x_scale
    y_scale = np.max(np.abs(ydata))
    ydata = ydata/y_scale
    
    x_range = np.max(xdata)-np.min(xdata)
    y_range = np.max(ydata)-np.min(ydata)
    x1_ini = np.min(xdata)+x_range/3
    x2_ini = np.max(xdata)-x_range/3
    x1_min = np.min(xdata)-x_range/4
    x1_max = np.max(xdata)-x_range/4
    x2_min = np.min(xdata)+x_range/4
    x2_max = np.max(xdata)+x_range/4
    
    def doubletanh(x, a, a0, w, x1, x2):
        """ This is a double hyperbolic tangent dip function. """
        y = a*(np.tanh(-(x-x1)/w)+1+np.tanh((x-x2)/w)+1)/2+a0
        return y
    
    # params: a,        a0,         w,          x1,     x2
    p_min = [y_range/3, 0,          x_range/10, x1_min, x2_min]
    p_ini = [y_range,   0,          x_range/4,  x1_ini, x2_ini]
    p_max = [y_range,   1-y_range/3,x_range/2,  x1_max, x2_max]
    
    try:
        popt, pcov = curve_fit(doubletanh, xdata, ydata, p0=p_ini, bounds=(p_min,p_max))
        perr = np.sqrt(np.diag(pcov)) # std dev, 2*perr is 95% conf int
    except RuntimeError as err:
        if handle_bad_fit == True:
            print(*err.args, '\n'+'Returning initial fit parameters instead!')
            popt = p_ini
            perr = np.tile(float('nan'), np.size(popt))
        else:
            raise
    
    # # Plot (for debugging)
    # plt.close(2)
    # plt.figure(2)
    # plt.plot(xdata, ydata, 'o')
    # plt.plot(xdata, doubletanh(xdata, *popt))
    # plt.xlabel('xdata/x_scale')
    # plt.ylabel('ydata/y_scale')
    # plt.show()
    
    # Normalize back
    for i in [0, 1]:
        popt[i] = popt[i]*y_scale
        perr[i] = perr[i]*y_scale
    for i in [2, 3, 4]:
        popt[i] = popt[i]*x_scale
        perr[i] = perr[i]*x_scale
    
    return popt, perr, doubletanh



def fit_fano(xdata, ydata, EXPONENT=2, handle_bad_fit=False):
    """ Fits the data using a Fano resonance lineshape. 
    
    Use the EXPONENT=1 option to fit the transmission data directly to the 
    sqrt(lorentzian). In this case fwhm is still the lorentzian FWHM, i.e. the 
    full width at 0.71*max.

    """
    # Normalization and initial values
    x_scale = 10**np.floor(np.log10(np.max(np.abs(xdata))))
    xdata = xdata/x_scale
    y_scale = np.max(np.abs(ydata))
    ydata = ydata/y_scale
        
    x_range = np.max(xdata)-np.min(xdata)
    y_range = np.max(ydata)-np.min(ydata)
        
    idx = np.argmin(np.abs(ydata-np.max(ydata)))
    f0_ini = np.take(xdata, idx)
    f0_min = np.min(xdata)
    f0_max = np.max(xdata)
        
    def fano_peak(x, f0, fwhm, a, q, r):
        """
        Fano resonance peak function (if EXPONENT=2). Can be used for Lorentzian 
        fit if q->Inf and r=0.

        Parameters
        ----------
        x : vector
            Frequency data.
        f0 : scalar
            Resonance frequency.
        fwhm : scalar
            Full width at half maximum.
        a : scalar
            Amplitude.
        q : scalar > 1
            Fano factor magnitude. When q -> Inf, the functions reduces to a 
            Lorentzian lineshape.
        r : scalar [0, pi]
            Fano factor phase in rad. For r=0, the resonance peak has a higher 
            transmission on the high frequency side. For r=pi, it is on the low 
            frequency side.

        Returns
        -------
        y : vector
            Transmission data.

        """
        y = a*np.abs( ( (x-f0)/(q*np.exp(1j*r))+fwhm/2 )/( 1j*(x-f0)+fwhm/2 ) )**EXPONENT
        return y
    
    # params: f0,       fwhm,       a,          q,          r
    p_min = [f0_min,    0,          0,          1,          0]
    p_ini = [f0_ini,    x_range/5,  y_range,    10,         np.pi/2]
    p_max = [f0_max,    x_range,    y_range,    1e5,        np.pi]
    
    try:
        popt, pcov = curve_fit(fano_peak, xdata, ydata, p0=p_ini, bounds=(p_min,p_max))
        perr = np.sqrt(np.diag(pcov)) # std dev, 2*perr is 95% conf int
    except RuntimeError as err:
        if handle_bad_fit == True:
            print(*err.args, '\n'+'Returning initial fit parameters instead!')
            popt = p_ini
            perr = np.tile(float('nan'), np.size(popt))
        else:
            raise
    
    # # Plot (for debugging)
    # plt.close(2)
    # plt.figure(2)
    # plt.plot(xdata, ydata, 'o')
    # plt.plot(xdata, fano_peak(xdata, *popt))
    # plt.xlabel('xdata/x_scale')
    # plt.ylabel('ydata/y_scale')
    # plt.show()
    
    # Normalize back
    for i in [2]:
        popt[i] = popt[i]*y_scale
        perr[i] = perr[i]*y_scale
    for i in [0, 1]:
        popt[i] = popt[i]*x_scale
        perr[i] = perr[i]*x_scale
    
    return popt, perr, fano_peak

def find_closest(array, value, axis=None):
    """ Finds the closest elements in array to value along an axis.

    Parameters
    ----------
    array : ndarray, list
        A 1D or 2D array or list.
    value : scalar
        A scalar value to compare.
    axis : None, 0, 1, optional
        Recommended to provide one for 2D array. The default is None.

    Returns
    -------
    out_array : ndarray, one less dimention than array
        Contains the closest values along the axis specified. Not necessarili along the same splice.
    idx : scalar or ndarray
        The indices of the closest elements along axis.

    """
    array = np.asarray(array)
    idx = np.argmin(np.abs(array-value), axis=axis)
    out_array = np.take(array, idx, axis=axis)
    if np.size(idx) > 1:
        out_array = np.diagonal(out_array)
    return out_array, idx


def calc_2dscanturbo_average(dataset_arr):
    Data_set = []

    for data_RT_IQ in dataset_arr:
        filename = str(data_RT_IQ) + "/frequency_RP2.dat"
        ds = pd.read_csv(filename, skiprows=[1,2], sep='\t').to_numpy()
        Data_set.append(ds)
    
    avg_ds  = np.mean(Data_set,axis=0)

    return avg_ds


# plt.style.use('default')
plt.style.use('seaborn-paper')
current_palette = sns.color_palette(palette=None)
sns.set_palette(palette=None)
# sns.palplot(current_palette)



mpl.rcParams.update({'figure.figsize': (8, 6 )})
mpl.rcParams['figure.dpi']= 200

SMALL_SIZE = 10*2
MEDIUM_SIZE = 12*2
BIGGER_SIZE = 14*2

plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
# plt.rcParams["font.weight"] = "bold"
plt.rcParams['axes.linewidth'] = 3         # set axes weight value globally
plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

# mpl.rcParams.update(mpl.rcParamsDefault)
tick_width = 2
tick_length = 10

plt.rcParams["xtick.major.width"] = tick_width
plt.rcParams["xtick.minor.width"] = tick_width
plt.rcParams["xtick.major.size"] = tick_length
plt.rcParams["xtick.minor.size"] = tick_length/2

plt.rcParams["ytick.major.width"] = tick_width
plt.rcParams["ytick.minor.width"] = tick_width
plt.rcParams["ytick.major.size"] = tick_length
plt.rcParams["ytick.minor.size"] = tick_length/2

# %% Universal Constant and Function Declaration
from scipy.constants import pi, h, hbar, elementary_charge, Boltzmann, speed_of_light

## Special Functions ##
from scipy.special.cython_special import ellipeinc


# %% Widescan

data_df = pd.read_csv("Widescan_RFscan2D/LP2_RP2.dat", skiprows=[1,2], sep='\t')

extent = [data_df['# LP2'].min(), data_df['# LP2'].max(), data_df['RP2'].min(), data_df['RP2'].max()]
cmap = mpl.cm.viridis

I =  np.array(data_df['I'])
Q =  np.array(data_df['Q'])
s21mag = np.sqrt(I**2 + Q**2)

S21_reshape = np.reshape(s21mag, (100, 100)) 


fig = plt.figure(figsize=(6, 5))
plt.imshow(S21_reshape, origin='upper', extent=extent, aspect = 'auto', cmap=cmap)
plt.colorbar(label="$V_{out}$ [V]", orientation="vertical")


plt.xlabel('$V_{RP} [mV]$')
plt.ylabel('$V_{LP}$ [mV]')

#%% Zoomed

data_df = pd.read_csv("Zoomed_RFscan2D/LP2_RP2.dat", skiprows=[1,2], sep='\t')

extent = [data_df['# LP2'].min(), data_df['# LP2'].max(), data_df['RP2'].min(), data_df['RP2'].max()]
cmap = mpl.cm.viridis

I =  np.array(data_df['I'])
Q =  np.array(data_df['Q'])
s21mag = np.sqrt(I**2 + Q**2)

S21_reshape = np.reshape(s21mag, (100, 100)) 

fig = plt.figure(figsize=(6, 5))
plt.imshow(S21_reshape, origin='upper', extent=extent, aspect = 'auto', cmap=cmap)
plt.colorbar(label="$V_{out}$ [V]", orientation="vertical")

plt.xlabel('$V_{RP} [mV]$')
plt.ylabel('$V_{LP}$ [mV]')

    