# Generate the Quantum capacitance values
# %% Import Packages Declaration

import matplotlib.pyplot as plt
import setuptools
import seaborn as sns
import matplotlib as mpl

import numpy as np
import pandas as pd
import scipy as sp
import skrf as rf
from qutip import *
from numpy import pi

from scipy.optimize import curve_fit
from colour import Color
from matplotlib.colors import LinearSegmentedColormap

from lmfit.models import LorentzianModel, GaussianModel, ConstantModel, LinearModel
from lmfit import fit_report
from lmfit import Model

from matplotlib.ticker import AutoMinorLocator


# %% Global plt.plot Setting 

# plt.style.use('default')
plt.style.use('seaborn-paper')
current_palette = sns.color_palette(palette=None)
sns.set_palette(palette=None)
# sns.palplot(current_palette)



mpl.rcParams.update({'figure.figsize': (8, 6 )})
mpl.rcParams['figure.dpi']= 200

SMALL_SIZE = 10*2
MEDIUM_SIZE = 12*2
BIGGER_SIZE = 14*2

plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
# plt.rcParams["font.weight"] = "bold"
plt.rcParams['axes.linewidth'] = 3         # set axes weight value globally
plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

# mpl.rcParams.update(mpl.rcParamsDefault)
tick_width = 2
tick_length = 10

plt.rcParams["xtick.major.width"] = tick_width
plt.rcParams["xtick.minor.width"] = tick_width
plt.rcParams["xtick.major.size"] = tick_length
plt.rcParams["xtick.minor.size"] = tick_length/2

plt.rcParams["ytick.major.width"] = tick_width
plt.rcParams["ytick.minor.width"] = tick_width
plt.rcParams["ytick.major.size"] = tick_length
plt.rcParams["ytick.minor.size"] = tick_length/2

# %% Universal Constant and Function Declaration

from scipy.constants import pi, h, hbar, elementary_charge, Boltzmann, speed_of_light

from scipy.special.cython_special import ellipeinc, ellipe


# %% Function List for script

def hamiltonian(eps, tc):
    """ 
    Hamiltonian of a 2 level system
    """
    H = (eps/2 * sigmaz() + tc * sigmax())
    
    return Qobj(H)

def hamiltonian_triplet(alpha, beta):
    """ 
    Hamiltonian for the triplet states
    """
    H = (alpha * sigmaz() + beta * sigmax())
    
    return Qobj(H)


def plot_energy(eps_vec, energies):
    """ 
    Energy level plotter
    """
    for n in range(len(energies[0,:])):
        plt.plot(eps_vec, energies[:,n], linewidth=4)
        plt.xlabel('DQD Detuning $\epsilon$ (GHz) ')
        plt.ylabel('Energy (GHz)')

    
def freq_to_meV(freq):
    meV = freq * h * 6.241506363094E+18 * 10**(3)

    return meV

def freq_to_ueV(freq):
    meV = freq * h * 6.241506363094E+18 * 10**(6)

    return meV

def gen_energy_diagram(eps_vec, B_z, B_x, B_tot, tc):
    """ 
    Hamiltonian 2 level system solver for a given detuning and tunnel coupling
    
    :param  eps_vec: detuning vector space [Hz]
            B_z, B_x, B_tot: Magnetic field *Depreciated need to rework [T]
            tc: tunneling coupling factor [Hz]
            
    :return energies vs detuning at ground and excited states
    """
    g = 2 # G factor 
    mu_b = 9.274 * 10**(-24) # Bohr Magnetron Constant
    B_tot = 0.15
    tc = tc*h * 6.241506363094E+18
    eps_vec = eps_vec*h* 6.241506363094E+18
    energies = np.array([hamiltonian(eps, tc).eigenenergies() for eps in eps_vec])
    energies_triplet = np.array([hamiltonian(g*B_z*mu_b, g*B_x*mu_b*0.8).eigenenergies() for eps in eps_vec])

    return energies #In energy

def calc_Cq(eps_vec, energies, beta, tc):
    """ 
    Function to calculate quantum capacitance given a energy diagram profile
    
    :param  eps_vec: detuning vector space
            energies: energy sets from hamiltonian solver
            beta: lever arm factor
            tc: tunneling energy
            
    :return Cq quantum capacitance [array]
    """
    eps_J = eps_vec*h # Convert ε detuning [Hz] to joules
    tc_J =  tc * h # Convert tc[Hz] to joules
    Cq = (beta*elementary_charge)**2 * (2*tc_J**2)/(eps_J**2 + 4*(tc_J)**2)**(3/2)

    return Cq
    
def calc_eff_Cq(eps_vec, energies, beta, tc, Vgate):
    """ 
    Function to calculate EFFECTIVE quantum capacitance given a input voltage to the gate
    Note that this is calculated using the analytical equation derived previously (See Companion PPTX)
    
    :param  eps_vec: detuning vector space
            energies: energy sets from hamiltonian solver
            beta: lever arm factor
            tc: tunneling energy
            Vgate: Voltage swing across the capacitance profile
            
    :return Cq_eff averaged over a voltage swing of N cycles [array]
    """
    eps_J = eps_vec*h # Convert ε detuning [Hz] to joules
    tc_J =  tc * h # Convert tc[Hz] to joules
    beta_eV = beta*elementary_charge
    Cq_eff = []
    for Vg in Vgate:
        Cq = (beta_eV**2 * tc_J * ellipeinc(2*pi, (-beta_eV**2 * Vg**2)/(4*tc_J**2)))/(2*(beta_eV**2*Vg**2 + 4*tc_J**2)*pi)
        Cq_eff.append(Cq)
    return np.array(Cq_eff)



    
def find_nearest(array, value):
    """ 
    Function to find nearest value of an array to a given dataset
    
    :param  array: input array
            value: lookup value
    
    :return array[idx]: closest value in the dataset
            idx: the index at which the data is located
    
    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx], idx




# %% Section: Relevant parameters for simulation
N_sample = 100

eps_vec = np.linspace(-25*10**(9), 25*10**(9), num = N_sample) # Detuning in [Hz]
eps_eV = freq_to_meV(eps_vec)*10**(3)

Zres=4500
#---- DQD Properties ----#
B_z = 0.0 # Tesla
B_x = 0.00 # Tesla magnetic x moment
B_tot = 0

tc = 14.1e9/2 
beta = 0.102

#---- Vdev vs SNR  ----# 1D ANALYSIS
scale = 100
N_sample = 500

Vgate_pos = np.logspace(-7, 3, num = int(N_sample)) # Voltage amplitude applied at gate of DQD [V]
Vgate = np.append(-1*Vgate_pos[::-1], Vgate_pos)

# %%  Instantaneous Quantum capacitance from energy level diagram 

vgate_vec = Vgate

eps_ev = vgate_vec * beta
eps_Hz = eps_ev/(h*6.241506363094E+18) # Convert detuning [eV] to detuning [Hz]


energies = gen_energy_diagram(eps_Hz, B_z, B_x, B_tot, tc)
Cq = calc_Cq(eps_Hz, energies, beta, tc)

Cq_db = np.transpose(np.array([vgate_vec, Cq]))


zeros = np.zeros(len(Cq))

t_arr = np.linspace(0,1, num = len(Vgate))

Cq_eff_manual = []


# %% Plot Cq

plt.figure(figsize=(7,4))
plt.plot(Vgate*beta*1e6, Cq_db[:,1]*1e18, linewidth=4, color='#2525fe', label='$|S\\rangle$' )
plt.axhline(y=0, color='k', linewidth=4, label='$|T\\rangle$')

plt.xlim([-150,150])
plt.ylim([-1, 20])
plt.ylabel('$C_{q}$ [aF]')
# plt.xlabel('$V_{dev} [V$]')
plt.xlabel('Detuning $\epsilon$ $[\mu eV]$')
plt.grid()
plt.legend()
    
plt.tight_layout()
# plt.savefig('Cq_Profile.png', format='png', dpi=600, transparent=True)

# %%Calculate Quantum Capacitance and Plot them

Cq_eff = calc_eff_Cq(eps_vec, energies, beta, tc, Vgate)

d = {'Vdev':Vgate, 'Cq_eff':Cq_eff} 
Cqeff_df = pd.DataFrame(data=d) # Convert Cqeff to a database library with corresponding gate voltage



# %% Plot Cqeff

plt.figure(figsize=(7,4))

plt.plot(Vgate*1e3, Cq_eff*1E18, linewidth=4, color='#2525fe', label='Equation')

plt.ylim([0,20])
plt.xlim([0,1])
plt.ylabel('$C_{q,eff}$ [aF]')
plt.xlabel('$|V_{A}|$ [mV]')

plt.grid()








