# Generate multiple Cqeff with different tc

# %% Import Packages Declaration

import matplotlib.pyplot as plt
import setuptools
import seaborn as sns
import matplotlib as mpl

import numpy as np
import pandas as pd
import scipy as sp
import skrf as rf
from qutip import *
from numpy import pi

from scipy.optimize import curve_fit
from colour import Color
from matplotlib.colors import LinearSegmentedColormap

from lmfit.models import LorentzianModel, GaussianModel, ConstantModel, LinearModel
from lmfit import fit_report
from lmfit import Model

from matplotlib.ticker import AutoMinorLocator


# %% Global plt.plot Setting 

# plt.style.use('default')
plt.style.use('seaborn-paper')
current_palette = sns.color_palette(palette=None)
sns.set_palette(palette=None)
# sns.palplot(current_palette)



mpl.rcParams.update({'figure.figsize': (8, 6 )})
mpl.rcParams['figure.dpi']= 200

SMALL_SIZE = 10*2
MEDIUM_SIZE = 12*2
BIGGER_SIZE = 14*2

plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
# plt.rcParams["font.weight"] = "bold"
plt.rcParams['axes.linewidth'] = 3         # set axes weight value globally
plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

# mpl.rcParams.update(mpl.rcParamsDefault)
tick_width = 2
tick_length = 10

plt.rcParams["xtick.major.width"] = tick_width
plt.rcParams["xtick.minor.width"] = tick_width
plt.rcParams["xtick.major.size"] = tick_length
plt.rcParams["xtick.minor.size"] = tick_length/2

plt.rcParams["ytick.major.width"] = tick_width
plt.rcParams["ytick.minor.width"] = tick_width
plt.rcParams["ytick.major.size"] = tick_length
plt.rcParams["ytick.minor.size"] = tick_length/2

# %% Universal Constant and Function Declaration

from scipy.constants import pi, h, hbar, elementary_charge, Boltzmann, speed_of_light

from scipy.special.cython_special import ellipeinc, ellipe


# %% Function List for script

def dbm2vp(val):
    vp = np.sqrt(2)*np.sqrt(dbm2p(val)*50)
    return vp
    
def db20(val):
    val_db20 = 20*np.log10(abs(val))
    return val_db20


def db10(val):
    val_db10 = 10*np.log10(abs(val))    
    return val_db10


def v2dbm(val):
    val_dbm = 10 + 20*np.log10(abs(val))
    return val_dbm

def p2dbm(val):
    val_dbm = 10*np.log10(val) + 30
    return val_dbm

def dbm2p(val):
    val_p = 10**(val/10)*0.001    
    return val_p

def hamiltonian(eps, tc):
    """ 
    Hamiltonian of a 2 level system
    """
    H = (eps/2 * sigmaz() + tc * sigmax())
    
    return Qobj(H)

def hamiltonian_triplet(alpha, beta):
    """ 
    Hamiltonian for the triplet states
    """
    H = (alpha * sigmaz() + beta * sigmax())
    
    return Qobj(H)


def plot_energy(eps_vec, energies):
    """ 
    Energy level plotter
    """
    for n in range(len(energies[0,:])):
        plt.plot(eps_vec, energies[:,n], linewidth=4)
        plt.xlabel('DQD Detuning $\epsilon$ (GHz) ')
        plt.ylabel('Energy (GHz)')

    
def freq_to_meV(freq):
    meV = freq * h * 6.241506363094E+18 * 10**(3)

    return meV

def freq_to_ueV(freq):
    meV = freq * h * 6.241506363094E+18 * 10**(6)

    return meV

def gen_energy_diagram(eps_vec, B_z, B_x, B_tot, tc):
    """ 
    Hamiltonian 2 level system solver for a given detuning and tunnel coupling
    
    :param  eps_vec: detuning vector space [Hz]
            B_z, B_x, B_tot: Magnetic field *Depreciated need to rework [T]
            tc: tunneling coupling factor [Hz]
            
    :return energies vs detuning at ground and excited states
    """
    g = 2 # G factor 
    mu_b = 9.274 * 10**(-24) # Bohr Magnetron Constant
    B_tot = 0.15
    tc = tc*h * 6.241506363094E+18
    eps_vec = eps_vec*h* 6.241506363094E+18
    energies = np.array([hamiltonian(eps, tc).eigenenergies() for eps in eps_vec])
    energies_triplet = np.array([hamiltonian(g*B_z*mu_b, g*B_x*mu_b*0.8).eigenenergies() for eps in eps_vec])
    
    # plot_energy(freq_to_meV(eps_vec)*10**(3), freq_to_meV(energies)*10**(3));
    # plot_energy(freq_to_meV(energies_triplet)*10**(3), freq_to_meV(energies)*10**(3));
    
    # plt.xlabel('Detuning $\epsilon$ (a.u)')
    # plt.ylabel('Energy (a.u)')
    # plt.grid()
    # E_T0 = freq_to_meV(eps_vec)*10**(3)/2;
    # E_T1 = freq_to_meV(eps_vec)*10**(3)/2 + g*B_tot*100
    # E_T2 = freq_to_meV(eps_vec)*10**(3)/2 - g*B_tot*100
    # plt.plot(freq_to_meV(eps_vec)*10**3, E_T0, color='black')
    # plt.plot(freq_to_meV(eps_vec)*10**3, E_T1, color='black')
    # plt.plot(freq_to_meV(eps_vec)*10**3, E_T2, color='black')
    # plt.ylim([-42,42])
    # plt.xlim([-80,80])
    
    # plt.tick_params(axis='both', labelsize=0, length = 0)


    return energies #In energy

def calc_Cq(eps_vec, energies, beta, tc):
    """ 
    Function to calculate quantum capacitance given a energy diagram profile
    
    :param  eps_vec: detuning vector space
            energies: energy sets from hamiltonian solver
            beta: lever arm factor
            tc: tunneling energy
            
    :return Cq quantum capacitance [array]
    """
    eps_J = eps_vec*h # Convert ε detuning [Hz] to joules
    tc_J =  tc * h # Convert tc[Hz] to joules
    Cq = (beta*elementary_charge)**2 * (2*tc_J**2)/(eps_J**2 + 4*(tc_J)**2)**(3/2)

    return Cq
    
def calc_eff_Cq(eps_vec, energies, beta, tc, Vgate):
    """ 
    Function to calculate EFFECTIVE quantum capacitance given a input voltage to the gate
    Note that this is calculated using the analytical equation derived previously (See Companion PPTX)
    
    :param  eps_vec: detuning vector space
            energies: energy sets from hamiltonian solver
            beta: lever arm factor
            tc: tunneling energy
            Vgate: Voltage swing across the capacitance profile
            
    :return Cq_eff averaged over a voltage swing of N cycles [array]
    """
    eps_J = eps_vec*h # Convert ε detuning [Hz] to joules
    tc_J =  tc * h # Convert tc[Hz] to joules
    beta_eV = beta*elementary_charge
    Cq_eff = []
    for Vg in Vgate:
        Cq = (beta_eV**2 * tc_J * ellipeinc(2*pi, (-beta_eV**2 * Vg**2)/(4*tc_J**2)))/(2*(beta_eV**2*Vg**2 + 4*tc_J**2)*pi)
        Cq_eff.append(Cq)
    return np.array(Cq_eff)




    
def find_nearest(array, value):
    """ 
    Function to find nearest value of an array to a given dataset
    
    :param  array: input array
            value: lookup value
    
    :return array[idx]: closest value in the dataset
            idx: the index at which the data is located
    
    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx], idx



def colored_line(x, y, z=None, line_width=1, MAP='jet'):
    # use pcolormesh to make interpolated rectangles
    num_pts = len(x)
    [xs, ys, zs] = [
        np.zeros((num_pts,2)),
        np.zeros((num_pts,2)),
        np.zeros((num_pts,2))
    ]

    dist = 0
    out_of_page = [0, 0, 1]
    for i in range(num_pts):
        # set the colors and the x,y locations of the source line
        xs[i][0] = x[i]
        ys[i][0] = y[i]
        if i > 0:
            x_delta =  x[i] - x[i-1]
            y_delta =  y[i] - y[i-1]
            seg_length = (x_delta**2 + y_delta**2)**0.5
            dist += seg_length
            zs[i] = [dist, dist]

        # define the offset perpendicular points
        if i == num_pts - 1:
            current_seg = [x[i]-x[i-1], y[i]-y[i-1], 0]
        else:
            current_seg = [x[i+1]-x[i], y[i+1]-y[i], 0]
        current_seg_perp = _get_perp_line(
            current_seg, out_of_page, line_width)
        if i == 0 or i == num_pts - 1:
            xs[i][1] = xs[i][0] + current_seg_perp[0]
            ys[i][1] = ys[i][0] + current_seg_perp[1]
            continue
        current_pt = [x[i], y[i]]
        current_seg_unit = _get_unit_vector(current_seg)
        previous_seg = [x[i]-x[i-1], y[i]-y[i-1], 0]
        previous_seg_perp = _get_perp_line(
            previous_seg, out_of_page, line_width)
        previous_seg_unit = _get_unit_vector(previous_seg)
        # current_pt + previous_seg_perp + scalar * previous_seg_unit =
        # current_pt + current_seg_perp - scalar * current_seg_unit =
        scalar = (
            (current_seg_perp - previous_seg_perp) /
            (previous_seg_unit + current_seg_unit)
        )
        new_pt = current_pt + previous_seg_perp + scalar[0] * previous_seg_unit
        xs[i][1] = new_pt[0]
        ys[i][1] = new_pt[1]

    # fig, ax = plt.subplots()
    # MAP = make_Ramp( ['#3776ab','black']) 
    # cm = plt.get_cmap(MAP)
    # ax.pcolormesh(xs, ys, zs, shading='gouraud', cmap=cm)
    return MAP,xs,ys,zs

def make_Ramp(ramp_colors): 
    from colour import Color
    from matplotlib.colors import LinearSegmentedColormap

    color_ramp = LinearSegmentedColormap.from_list( 'my_list', [ Color( c1 ).rgb for c1 in ramp_colors ] )
    plt.figure( figsize = (15,3))
    plt.imshow( [list(np.arange(0, len( ramp_colors ) , 0.1)) ] , interpolation='nearest', origin='lower', cmap= color_ramp )
    plt.xticks([])
    plt.yticks([])
    return color_ramp



def rotate_matrix (x, y, angle, x_shift=0, y_shift=0, units="DEGREES"):
    """
    Rotates a point in the xy-plane counterclockwise through an angle about the origin
    https://en.wikipedia.org/wiki/Rotation_matrix
    :param x: x coordinate
    :param y: y coordinate
    :param x_shift: x-axis shift from origin (0, 0)
    :param y_shift: y-axis shift from origin (0, 0)
    :param angle: The rotation angle in degrees
    :param units: DEGREES (default) or RADIANS
    :return: Tuple of rotated x and y
    """

    # Shift to origin (0,0)
    x = x - x_shift
    y = y - y_shift

    # Convert degrees to radians
    if units == "DEGREES":
        angle = np.radians(angle)

    # Rotation matrix multiplication to get rotated x & y
    xr = (x * np.cos(angle)) - (y * np.sin(angle)) + x_shift
    yr = (x * np.sin(angle)) + (y * np.cos(angle)) + y_shift

    return xr, yr



# %% Section: Relevant parameters for simulation
N_sample = 100

eps_vec = np.linspace(-25*10**(9), 25*10**(9), num = N_sample) # Detuning in [Hz]
eps_eV = freq_to_meV(eps_vec)*10**(3)

Zres=4500
#---- DQD Properties ----#
B_z = 0.0 # Tesla
B_x = 0.00 # Tesla magnetic x moment
B_tot = 0

tc_factor = np.arange(2,11,2)
plt.figure(figsize=(6,4))

for factor in tc_factor:
        
    tc = 6.9e9/2 * factor # Tc sweep
    
    beta = 0.102
    scale = 100
    N_sample = 500

    Vgate = np.linspace(-5e-3, 5e-3, num = N_sample) # Voltage amplitude applied at gate of DQD [V]
            
    N_sample_eps = N_sample
    A_arr = Vgate
    vgate_vec = np.linspace(-max(A_arr), max(A_arr), num = N_sample)    
    eps_ev = vgate_vec * beta
    eps_Hz = eps_ev/(h*6.241506363094E+18) # Convert detuning [eV] to detuning [Hz]
    
    
    energies = gen_energy_diagram(eps_Hz, B_z, B_x, B_tot, tc)
    Cq = calc_Cq(eps_Hz, energies, beta, tc)
    
    Cq_db = np.transpose(np.array([vgate_vec, Cq]))
    

    zeros = np.zeros(len(Cq))
    
    t_arr = np.linspace(0,1, num = N_sample)
    
    Cq_eff_manual = []
    
    for A in Vgate:
        Cq_proj = []
        Vgate_t = A*np.sin((2*pi)*t_arr)
    
        for Vgate_var in Vgate_t:
            val, idx = find_nearest(Cq_db[:,0], Vgate_var)
            Cq_proj.append(Cq_db[idx,1])
    
        Cq_eff_manual.append(np.sum(Cq_proj)/len(Cq_proj))

    
    Cq_eff = calc_eff_Cq(eps_vec, energies, beta, tc, Vgate)
    
    d = {'Vdev':Vgate, 'Cq_eff':Cq_eff} 
    Cqeff_df = pd.DataFrame(data=d) 
    
    err = 1 - Cq_eff/Cq_eff_manual
    
    plt.plot(Vgate*1e3, Cq_eff*1E18, linewidth=2,  label = '$(2t_c/h)/f_r$='+str(factor))
    
plt.xlim([0,1])
plt.ylabel('$C_{q,eff}$ [aF]')
plt.xlabel('$|V_{A}| [m$V]')

plt.grid()
    
plt.legend(loc=1, prop={'size': 15})






