
# %% Import Declaration
import matplotlib.pyplot as plt
import setuptools
import seaborn as sns
import matplotlib as mpl
from matplotlib.collections import LineCollection
from matplotlib.colors import ListedColormap, BoundaryNorm


import json
import os


import numpy as np
import pandas as pd
import scipy as sp
import skrf as rf
from qutip import *
from numpy import pi

from scipy.optimize import curve_fit
from scipy import special as sp
from pathlib import Path
from lmfit.models import LorentzianModel, gaussian, ConstantModel, LinearModel
from lmfit import fit_report

import warnings
warnings.filterwarnings("ignore")
# %% Global plt.plot Setting 

# plt.style.use('default')
plt.style.use('seaborn-paper')
current_palette = sns.color_palette(palette=None)
sns.set_palette(palette=None)
# sns.palplot(current_palette)



mpl.rcParams.update({'figure.figsize': (8, 6 )})
mpl.rcParams['figure.dpi']= 200

SMALL_SIZE = 10*2
MEDIUM_SIZE = 12*2
BIGGER_SIZE = 14*2

plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
# plt.rcParams["font.weight"] = "bold"
plt.rcParams['axes.linewidth'] = 3         # set axes weight value globally
plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title


tick_width = 2
tick_length = 10

plt.rcParams["xtick.major.width"] = tick_width
plt.rcParams["xtick.minor.width"] = tick_width
plt.rcParams["xtick.major.size"] = tick_length
plt.rcParams["xtick.minor.size"] = tick_length/2

plt.rcParams["ytick.major.width"] = tick_width
plt.rcParams["ytick.minor.width"] = tick_width
plt.rcParams["ytick.major.size"] = tick_length
plt.rcParams["ytick.minor.size"] = tick_length/2


# mpl.rcParams.update(mpl.rcParamsDefault)
from scipy.constants import pi, h, hbar, elementary_charge, Boltzmann, speed_of_light

# %% Function def

def SNR_calc(s21mag, RP, A_custom , Custom_A=False):
    """ 
    Take s21mag data and fit with a guess A 'amplitude'. EN = True if you wish to put estimated amplitude at higher int time
    Output = [SNR_lin, SNR_dB, A, N_floor_std, Max_sig, Min_sig
    """   
    
    s21mag = np.array(s21mag)
    RP = np.array(RP)
    
    peak = LorentzianModel() 
    offset = ConstantModel()
    model = peak + offset
    
    pars = offset.make_params(c=np.mean(s21mag))

    

    pars += peak.guess(s21mag, x=RP, amplitude= -1)
    # pars += peak.guess(data_df['Q'], x=data_df['# RP2'], amplitude=-1)

    result = model.fit(s21mag, pars, x=RP)
    # result = model.fit(data_df['Q'], pars, x=data_df['# RP2'])

    if Custom_A == False:
        A = result.params['height'].value 
    if Custom_A == True:
        A = A_custom
        
    Max_sig = result.params['c'].value
    Min_sig = Max_sig - abs(A)
    N_floor_std = np.std(s21mag[-20:-1] - result.params['c'])
    # N_floor_std = np.std(data_df['S21mag'][0:100] - result.params['c'])

    SNR_lin = (A**2/N_floor_std**2)
    SNR_dB = 10*np.log10(SNR_lin)


    # print(min(data_df['S21mag']))
    # print(max(data_df['S21mag']))

    # plt.figure()
    plt.plot(RP, result.best_fit, color='k')
    plt.plot(RP, s21mag,"-")
    plt.xlabel('$V_{RP}$ [V]')
    
    return [SNR_lin, SNR_dB, A, N_floor_std, Max_sig, Min_sig]


def v2dbm(val):
    val_dbm = 10 + 20*np.log10(abs(val))
    return val_dbm


def fit_lorentzian(xdata, ydata):
    
    xdata = np.asarray(xdata)
    ydata = np.asarray(ydata)
    
    guess_amp = -1*(np.average(ydata) - min(ydata)) #initial amplitude guess
    
    peak = LorentzianModel() 
    offset = ConstantModel()
    model = peak + offset
    
    pars = offset.make_params(c=np.mean(ydata))

    pars += peak.guess(ydata, x=xdata, amplitude= guess_amp)
    # pars += peak.guess(data_df['Q'], x=data_df['# RP2'], amplitude=-1)

    result = model.fit(ydata, pars, x=xdata)
    # result = model.fit(data_df['Q'], pars, x=data_df['# RP2'])
    
    # plt.plot(xdata, result.best_fit, color='k')
    # plt.plot(xdata, ydata,"-")
    return result
    

def fit_doubletanh(xdata, ydata, handle_bad_fit=False):
    """ Fits the data using a symmetric double hyperbolic tangent function to 
    find the zero-detuning point of a dispersive shift vs detuning dataset.

    """
    # Normalization and initial values
    x_scale = 10**np.floor(np.log10(np.max(np.abs(xdata))))
    xdata = xdata/x_scale
    y_scale = np.max(np.abs(ydata))
    ydata = ydata/y_scale
    
    x_range = np.max(xdata)-np.min(xdata)
    y_range = np.max(ydata)-np.min(ydata)
    x1_ini = np.min(xdata)+x_range/3
    x2_ini = np.max(xdata)-x_range/3
    x1_min = np.min(xdata)-x_range/4
    x1_max = np.max(xdata)-x_range/4
    x2_min = np.min(xdata)+x_range/4
    x2_max = np.max(xdata)+x_range/4
    
    def doubletanh(x, a, a0, w, x1, x2):
        """ This is a double hyperbolic tangent dip function. """
        y = a*(np.tanh(-(x-x1)/w)+1+np.tanh((x-x2)/w)+1)/2+a0
        return y
    
    # params: a,        a0,         w,          x1,     x2
    p_min = [y_range/3, 0,          x_range/10, x1_min, x2_min]
    p_ini = [y_range,   0,          x_range/4,  x1_ini, x2_ini]
    p_max = [y_range,   1-y_range/3,x_range/2,  x1_max, x2_max]
    
    try:
        popt, pcov = curve_fit(doubletanh, xdata, ydata, p0=p_ini, bounds=(p_min,p_max))
        perr = np.sqrt(np.diag(pcov)) # std dev, 2*perr is 95% conf int
    except RuntimeError as err:
        if handle_bad_fit == True:
            print(*err.args, '\n'+'Returning initial fit parameters instead!')
            popt = p_ini
            perr = np.tile(float('nan'), np.size(popt))
        else:
            raise
    
    # # Plot (for debugging)
    # plt.close(2)
    # plt.figure(2)
    # plt.plot(xdata, ydata, 'o')
    # plt.plot(xdata, doubletanh(xdata, *popt))
    # plt.xlabel('xdata/x_scale')
    # plt.ylabel('ydata/y_scale')
    # plt.show()
    
    # Normalize back
    for i in [0, 1]:
        popt[i] = popt[i]*y_scale
        perr[i] = perr[i]*y_scale
    for i in [2, 3, 4]:
        popt[i] = popt[i]*x_scale
        perr[i] = perr[i]*x_scale
    
    return popt, perr, doubletanh



def fit_fano(xdata, ydata, EXPONENT=2, handle_bad_fit=False):
    """ Fits the data using a Fano resonance lineshape. 
    
    Use the EXPONENT=1 option to fit the transmission data directly to the 
    sqrt(lorentzian). In this case fwhm is still the lorentzian FWHM, i.e. the 
    full width at 0.71*max.

    """
    # Normalization and initial values
    x_scale = 10**np.floor(np.log10(np.max(np.abs(xdata))))
    xdata = xdata/x_scale
    y_scale = np.max(np.abs(ydata))
    ydata = ydata/y_scale
        
    x_range = np.max(xdata)-np.min(xdata)
    y_range = np.max(ydata)-np.min(ydata)
        
    idx = np.argmin(np.abs(ydata-np.max(ydata)))
    f0_ini = np.take(xdata, idx)
    f0_min = np.min(xdata)
    f0_max = np.max(xdata)
        
    def fano_peak(x, f0, fwhm, a, q, r):
        """
        Fano resonance peak function (if EXPONENT=2). Can be used for Lorentzian 
        fit if q->Inf and r=0.

        Parameters
        ----------
        x : vector
            Frequency data.
        f0 : scalar
            Resonance frequency.
        fwhm : scalar
            Full width at half maximum.
        a : scalar
            Amplitude.
        q : scalar > 1
            Fano factor magnitude. When q -> Inf, the functions reduces to a 
            Lorentzian lineshape.
        r : scalar [0, pi]
            Fano factor phase in rad. For r=0, the resonance peak has a higher 
            transmission on the high frequency side. For r=pi, it is on the low 
            frequency side.

        Returns
        -------
        y : vector
            Transmission data.

        """
        y = a*np.abs( ( (x-f0)/(q*np.exp(1j*r))+fwhm/2 )/( 1j*(x-f0)+fwhm/2 ) )**EXPONENT
        return y
    
    # params: f0,       fwhm,       a,          q,          r
    p_min = [f0_min,    0,          0,          1,          0]
    p_ini = [f0_ini,    x_range/5,  y_range,    10,         np.pi/2]
    p_max = [f0_max,    x_range,    y_range,    1e5,        np.pi]
    
    try:
        popt, pcov = curve_fit(fano_peak, xdata, ydata, p0=p_ini, bounds=(p_min,p_max))
        perr = np.sqrt(np.diag(pcov)) # std dev, 2*perr is 95% conf int
    except RuntimeError as err:
        if handle_bad_fit == True:
            print(*err.args, '\n'+'Returning initial fit parameters instead!')
            popt = p_ini
            perr = np.tile(float('nan'), np.size(popt))
        else:
            raise
    
    # # Plot (for debugging)
    # plt.close(2)
    # plt.figure(2)
    # plt.plot(xdata, ydata, 'o')
    plt.plot(xdata, fano_peak(xdata, *popt))
    # plt.xlabel('xdata/x_scale')
    # plt.ylabel('ydata/y_scale')
    # plt.show()
    
    # Normalize back
    for i in [2]:
        popt[i] = popt[i]*y_scale
        perr[i] = perr[i]*y_scale
    for i in [0, 1]:
        popt[i] = popt[i]*x_scale
        perr[i] = perr[i]*x_scale
    
    return popt, perr, fano_peak

def find_closest(array, value, axis=None):
    """ Finds the closest elements in array to value along an axis.

    Parameters
    ----------
    array : ndarray, list
        A 1D or 2D array or list.
    value : scalar
        A scalar value to compare.
    axis : None, 0, 1, optional
        Recommended to provide one for 2D array. The default is None.

    Returns
    -------
    out_array : ndarray, one less dimention than array
        Contains the closest values along the axis specified. Not necessarili along the same splice.
    idx : scalar or ndarray
        The indices of the closest elements along axis.

    """
    array = np.asarray(array)
    idx = np.argmin(np.abs(array-value), axis=axis)
    out_array = np.take(array, idx, axis=axis)
    if np.size(idx) > 1:
        out_array = np.diagonal(out_array)
    return out_array, idx


# %% Plot IQ

IQ = np.load("IQ_resp_1ms.npy")
segments_S = np.load("line_segment_S.npy")
segments_T = np.load("line_segment_T.npy")

G_RX = 10**(91.5/20)

RF_Power_meas = np.arange(-50, -12, 2) - 75
RF_Power_sim = np.arange(-50, 2, 2) - 75

fig, ax = plt.subplots(figsize=(7,5))
lc = LineCollection(segments_S)
lc2 = LineCollection(segments_T)

lc.set_array(RF_Power_sim)
lc.set_linewidth(3)
line = ax.add_collection(lc)

lc2.set_array(RF_Power_sim)
lc2.set_linewidth(3)
line2 = ax.add_collection(lc2)


plt.scatter(IQ[:,2]/G_RX, IQ[:,3]/G_RX, marker='v', c=RF_Power_meas, label='Meas. $(\epsilon \\neq0)$', s=75)
plt.clim(RF_Power_sim.min(), RF_Power_sim.max()) 

plt.scatter(IQ[:,0]/G_RX, IQ[:,1]/G_RX, marker='s', c=RF_Power_meas, label='Meas. $(\epsilon =0)$', s=75)

plt.clim(RF_Power_sim.min(), RF_Power_sim.max()) 


ax.set_xlabel('$I_{RF,out}$ [V]')
ax.set_ylabel('$Q_{RF,out}$ [V]')

ax.legend(prop={'size': 20})

ax.grid()
fig.colorbar(line,ax=ax, label="$P_{RF}$ [dBm]")
ax.set_xscale('log')
ax.set_xlim([1e-9,1e-5])
ax.autoscale_view()




