"""

"""
# %% Import Declaration
import matplotlib.pyplot as plt
import setuptools
import seaborn as sns
import matplotlib as mpl

import json
import os


import numpy as np
import pandas as pd
import scipy as sp
import skrf as rf
from qutip import *
from numpy import pi

from scipy.optimize import curve_fit
from scipy import special as sp
from pathlib import Path
from lmfit.models import LorentzianModel, gaussian, ConstantModel, LinearModel
from lmfit import fit_report

import warnings
warnings.filterwarnings("ignore")
# %% Global plt.plot Setting 

# plt.style.use('default')
plt.style.use('seaborn-paper')
current_palette = sns.color_palette(palette=None)
sns.set_palette(palette=None)
# sns.palplot(current_palette)

mpl.rcParams.update({'figure.figsize': (8, 6)})
mpl.rcParams['figure.dpi']= 250

SMALL_SIZE = 10*2
MEDIUM_SIZE = 12*2
BIGGER_SIZE = 14*2

plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
# plt.rcParams["font.weight"] = "bold"
plt.rcParams['axes.linewidth'] = 3         # set axes weight value globally
plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title


tick_width = 2
tick_length = 10

plt.rcParams["xtick.major.width"] = tick_width
plt.rcParams["xtick.minor.width"] = tick_width
plt.rcParams["xtick.major.size"] = tick_length
plt.rcParams["xtick.minor.size"] = tick_length/2

plt.rcParams["ytick.major.width"] = tick_width
plt.rcParams["ytick.minor.width"] = tick_width
plt.rcParams["ytick.major.size"] = tick_length
plt.rcParams["ytick.minor.size"] = tick_length/2


# mpl.rcParams.update(mpl.rcParamsDefault)
from scipy.constants import pi, h, hbar, elementary_charge, Boltzmann, speed_of_light

# %% Function def

def v2dbm(val):
    val_dbm = 10 + 20*np.log10(abs(val))
    return val_dbm


def fit_lorentzian(xdata, ydata):
    
    xdata = np.asarray(xdata)
    ydata = np.asarray(ydata)
    
    guess_amp = -1*(np.average(ydata) - min(ydata)) #initial amplitude guess
    
    peak = LorentzianModel() 
    offset = ConstantModel()
    model = peak + offset
    
    pars = offset.make_params(c=np.mean(ydata))

    pars += peak.guess(ydata, x=xdata, amplitude= guess_amp)

    result = model.fit(ydata, pars, x=xdata)

    return result
    

def fit_doubletanh(xdata, ydata, handle_bad_fit=False):
    """ Fits the data using a symmetric double hyperbolic tangent function to 
    find the zero-detuning point of a dispersive shift vs detuning dataset.

    """
    # Normalization and initial values
    x_scale = 10**np.floor(np.log10(np.max(np.abs(xdata))))
    xdata = xdata/x_scale
    y_scale = np.max(np.abs(ydata))
    ydata = ydata/y_scale
    
    x_range = np.max(xdata)-np.min(xdata)
    y_range = np.max(ydata)-np.min(ydata)
    x1_ini = np.min(xdata)+x_range/3
    x2_ini = np.max(xdata)-x_range/3
    x1_min = np.min(xdata)-x_range/4
    x1_max = np.max(xdata)-x_range/4
    x2_min = np.min(xdata)+x_range/4
    x2_max = np.max(xdata)+x_range/4
    
    def doubletanh(x, a, a0, w, x1, x2):
        """ This is a double hyperbolic tangent dip function. """
        y = a*(np.tanh(-(x-x1)/w)+1+np.tanh((x-x2)/w)+1)/2+a0
        return y
    
    # params: a,        a0,         w,          x1,     x2
    p_min = [y_range/3, 0,          x_range/10, x1_min, x2_min]
    p_ini = [y_range,   0,          x_range/4,  x1_ini, x2_ini]
    p_max = [y_range,   1-y_range/3,x_range/2,  x1_max, x2_max]
    
    try:
        popt, pcov = curve_fit(doubletanh, xdata, ydata, p0=p_ini, bounds=(p_min,p_max))
        perr = np.sqrt(np.diag(pcov)) # std dev, 2*perr is 95% conf int
    except RuntimeError as err:
        if handle_bad_fit == True:
            print(*err.args, '\n'+'Returning initial fit parameters instead!')
            popt = p_ini
            perr = np.tile(float('nan'), np.size(popt))
        else:
            raise

    
    # Normalize back
    for i in [0, 1]:
        popt[i] = popt[i]*y_scale
        perr[i] = perr[i]*y_scale
    for i in [2, 3, 4]:
        popt[i] = popt[i]*x_scale
        perr[i] = perr[i]*x_scale
    
    return popt, perr, doubletanh



def fit_fano(xdata, ydata, EXPONENT=2, handle_bad_fit=False):
    """ Fits the data using a Fano resonance lineshape. 
    
    Use the EXPONENT=1 option to fit the transmission data directly to the 
    sqrt(lorentzian). In this case fwhm is still the lorentzian FWHM, i.e. the 
    full width at 0.71*max.

    """
    # Normalization and initial values
    x_scale = 10**np.floor(np.log10(np.max(np.abs(xdata))))
    xdata = xdata/x_scale
    y_scale = np.max(np.abs(ydata))
    ydata = ydata/y_scale
        
    x_range = np.max(xdata)-np.min(xdata)
    y_range = np.max(ydata)-np.min(ydata)
        
    idx = np.argmin(np.abs(ydata-np.max(ydata)))
    f0_ini = np.take(xdata, idx)
    f0_min = np.min(xdata)
    f0_max = np.max(xdata)
        
    def fano_peak(x, f0, fwhm, a, q, r):
        """
        Fano resonance peak function (if EXPONENT=2). Can be used for Lorentzian 
        fit if q->Inf and r=0.

        Parameters
        ----------
        x : vector
            Frequency data.
        f0 : scalar
            Resonance frequency.
        fwhm : scalar
            Full width at half maximum.
        a : scalar
            Amplitude.
        q : scalar > 1
            Fano factor magnitude. When q -> Inf, the functions reduces to a 
            Lorentzian lineshape.
        r : scalar [0, pi]
            Fano factor phase in rad. For r=0, the resonance peak has a higher 
            transmission on the high frequency side. For r=pi, it is on the low 
            frequency side.

        Returns
        -------
        y : vector
            Transmission data.

        """
        y = a*np.abs( ( (x-f0)/(q*np.exp(1j*r))+fwhm/2 )/( 1j*(x-f0)+fwhm/2 ) )**EXPONENT
        return y
    
    # params: f0,       fwhm,       a,          q,          r
    p_min = [f0_min,    0,          0,          1,          0]
    p_ini = [f0_ini,    x_range/5,  y_range,    10,         np.pi/2]
    p_max = [f0_max,    x_range,    y_range,    1e5,        np.pi]
    
    try:
        popt, pcov = curve_fit(fano_peak, xdata, ydata, p0=p_ini, bounds=(p_min,p_max))
        perr = np.sqrt(np.diag(pcov)) # std dev, 2*perr is 95% conf int
    except RuntimeError as err:
        if handle_bad_fit == True:
            print(*err.args, '\n'+'Returning initial fit parameters instead!')
            popt = p_ini
            perr = np.tile(float('nan'), np.size(popt))
        else:
            raise
    
    # # Plot (for debugging)
    # plt.close(2)
    # plt.figure(2)
    # plt.plot(xdata, ydata, 'o')
    plt.plot(xdata, fano_peak(xdata, *popt))
    # plt.xlabel('xdata/x_scale')
    # plt.ylabel('ydata/y_scale')
    # plt.show()
    
    # Normalize back
    for i in [2]:
        popt[i] = popt[i]*y_scale
        perr[i] = perr[i]*y_scale
    for i in [0, 1]:
        popt[i] = popt[i]*x_scale
        perr[i] = perr[i]*x_scale
    
    return popt, perr, fano_peak

def find_closest(array, value, axis=None):
    """ Finds the closest elements in array to value along an axis.

    Parameters
    ----------
    array : ndarray, list
        A 1D or 2D array or list.
    value : scalar
        A scalar value to compare.
    axis : None, 0, 1, optional
        Recommended to provide one for 2D array. The default is None.

    Returns
    -------
    out_array : ndarray, one less dimention than array
        Contains the closest values along the axis specified. Not necessarili along the same splice.
    idx : scalar or ndarray
        The indices of the closest elements along axis.

    """
    array = np.asarray(array)
    idx = np.argmin(np.abs(array-value), axis=axis)
    out_array = np.take(array, idx, axis=axis)
    if np.size(idx) > 1:
        out_array = np.diagonal(out_array)
    return out_array, idx

def extract_Pin_vs_data(data,t_int_idx, Pin_extract,RP):
    s21mag = []
    s21phase = []
    IQ_data = []
    
    for i in range(len(Pin_extract)): # n = 
        # print(i)
        idx = (i*11) + t_int_idx
        
        s21_tmp = []
        phase_tmp = []
        IQ_tmp = []
        
        for j in range(len(data[idx][0][:])):
            
            s21_tmp.append(data[idx][0][j][0]) # for s21mag
            phase_tmp.append(data[idx][0][j][1]) # for s21phase
            IQ_tmp.append(complex(data[idx][0][j][2],data[idx][0][j][3])) # for complex I + jQ
        
   
        phase_offset = phase_tmp[63] # phase at [63] is our measured off-detuning
    
    
        phase_tmp = np.unwrap([x - phase_offset for x in phase_tmp], period=360)
    
        for k in range(len(RP)):
            re = s21_tmp[k]*np.cos(np.deg2rad(phase_tmp[k]))    
            im = s21_tmp[k]*np.sin(np.deg2rad(phase_tmp[k]))
            IQ_tmp[k] = complex(re,im)
            

        s21phase.append(phase_tmp)
        s21mag.append(s21_tmp)
        IQ_data.append(IQ_tmp)
    return s21mag, s21phase, IQ_data

def calc_SNR(RF_Power, RP, s21mag, IQ_data):
    IQ_interdot = []
    IQ_offres = []
    interdot_RP = []
    IQ_point_S = []
    IQ_point_T = []
    Phase_S = []
        
    # plt.figure(figsize=(5,4))
    for i in range(len(RF_Power)):
        result = fit_lorentzian(RP, np.asarray(s21mag[i]))
        dip_idx = find_closest(RP, result.values['center']) 
        dip_idx = dip_idx[1] 
        
        interdot_RP.append(dip_idx)
        
        IQ_interdot.append(IQ_data[i][dip_idx])
        IQ_offres.append(IQ_data[i][63])
        
        #Plot I/Q
        # plt.scatter(np.real(IQ_data[i][dip_idx]), np.imag(IQ_data[i][dip_idx]),color='k', s=50) 
        # plt.scatter(np.real(IQ_data[i][63]), np.imag(IQ_data[i][63]),color='r', s=50) 
    
    
        IQ_point_S.append(complex(np.real(IQ_data[i][dip_idx]), np.imag(IQ_data[i][dip_idx])))
        IQ_point_T.append(complex(np.real(IQ_data[i][63]), np.imag(IQ_data[i][63])))
              
        # plt.xscale('log')
    # plt.ylabel('Q [V]')
    # plt.xlabel('I [V]')
    
    Phase_S.append(np.angle(np.real(IQ_point_S) + 1j*np.imag(IQ_point_S), deg=True))
    
    A = abs(np.subtract(IQ_offres, IQ_interdot))**2 #V^2
    Phase_S.append(np.angle(np.subtract(IQ_offres, IQ_interdot)))
    
    N = []

    for i in range(len(s21mag)):
        N.append(calc_Noise(RP, s21mag[i]))
    
    N = np.square(N) # N in s21mag

    return A, N, Phase_S



def calc_Noise(RP, s21mag):
    
    s21mag = np.array(s21mag)
    RP = np.array(RP)
    
    peak = LorentzianModel() 
    offset = ConstantModel()
    model = peak + offset
    
    pars = offset.make_params(c=np.mean(s21mag))

    pars += peak.guess(s21mag, x=RP, amplitude= -1)

    result = model.fit(s21mag, pars, x=RP)

    lwr_lim = -14
    hgh_lim = -1
    N_floor_std = np.std(s21mag[lwr_lim:hgh_lim] - result.params['c']) # standard deviation of the noise sampling in off detuning

    return N_floor_std

# %% Grab SNR dataset 

dir_name = []

Set = 'Set1' # 

list_of_files = os.listdir(os.getcwd() + '/'+Set) #list of files in the current directory
for each_file in list_of_files:
    if each_file.endswith('.json'):  
        dir_name.append(each_file)

np.sort(dir_name) # re-order the naming scheme 
dir_name_HP = dir_name[0:142] # High power
dir_name_LP = dir_name[143:241] # Low power

data =  []
for i in range(len(dir_name)):
    with open(str(Set)+'/'+dir_name[i]) as f:
        # print(f)
        tmp1 = json.load(f)
        data.append(tmp1['Cal DQD2 before'])


RF_Power_HP = np.arange(-36,-10,+2)
RF_Power_LP = np.arange(-50,-34,+2)
RF_Power = np.arange(-50, -10, 2)
t_int = np.flip(4*1E-6*np.array([   1,    2,    4,    8,   16,   32,   64,  128,  256,  512, 1024]))  # Tint steps in measurement

s21mag = []
s21phase = []
IQ_data = []

RP_beta = 0.151 # RP lever arm converting value to voltage
RP_center = -590 # Center RP
RP = np.linspace(-200,200, 64)# Plunger linspace, 64 points




# %% Grab dataset Tint for 8us,128us, and 1024us
s21mag_8us, s21phase_8us, IQ_8us  = extract_Pin_vs_data(data, 9, RF_Power, RP)
s21mag_128us, s21phase_128us, IQ_128us  = extract_Pin_vs_data(data, 5, RF_Power, RP)
s21mag_1ms, s21phase_1ms, IQ_1ms  = extract_Pin_vs_data(data, 2, RF_Power, RP)

s21mag_8us, s21phase_8us, IQ_8us  = extract_Pin_vs_data(data, 9, RF_Power, RP)
s21mag_128us, s21phase_128us, IQ_128us  = extract_Pin_vs_data(data, 5, RF_Power, RP)
s21mag_1ms, s21phase_1ms, IQ_1ms  = extract_Pin_vs_data(data, 2, RF_Power, RP)


# %% Re-arrange datatsets low and high power

def rearange_data(s21mag, s21phase, IQ):
    s21mag_new = []
    s21phase_new = []
    IQ_new = []
    n = 1
    s21mag_new      = s21mag[12+n:] + s21mag[0:12+n] 
    s21phase_new    = s21phase[12+n:] + s21phase[0:12+n]
    IQ_new          = IQ[12+n:] + IQ[0:12+n]

    return s21mag_new, s21phase_new, IQ_new
    
s21mag_8us, s21phase_8us, IQ_8us  = rearange_data(s21mag_8us, s21phase_8us, IQ_8us)
s21mag_128us, s21phase_128us, IQ_128us  = rearange_data(s21mag_128us, s21phase_128us, IQ_128us )
s21mag_1ms, s21phase_1ms, IQ_1ms  = rearange_data(s21mag_1ms, s21phase_1ms, IQ_1ms)


# %% Load sim

Phase_ADS_sim = np.loadtxt("ADS_sim\Phase_diff.txt", delimiter='\t', skiprows=1)
A_ADS_sim = np.loadtxt("ADS_sim\ADS_sim_A.txt", delimiter='\t', skiprows=1)


# %% Amplitude vs Pin Plotter
G_RX = 10**((91.5)/10) # Gain from RT to sample 

A_1, N_1, Phase_1 = calc_SNR(RF_Power, RP, s21mag_8us, IQ_8us)
A_2, N_2, Phase_2 = calc_SNR(RF_Power, RP, s21mag_128us, IQ_128us)
A_3, N_3, Phase_3 = calc_SNR(RF_Power, RP, s21mag_1ms, IQ_1ms)


# %% Plot magnitdue Readout Signal 

plt.figure(figsize=(8,4)) 
plt.semilogy(A_ADS_sim[:,0], A_ADS_sim[:,1], '--', color='k', linewidth='4', label='Sim.')
plt.semilogy(RF_Power[:-1]-75, np.sqrt(A_3[:-1]/G_RX), '-o', linewidth='3', label='$T_{int}=1024\mu$s', markersize=8)

plt.xlim([-125,-80])
plt.ylim([1e-8,1e-6])
plt.grid()
plt.ylabel('$|A_{sig}|$ $[V]$') 
plt.xlabel('$P_{RF}$ [dBm]')
plt.legend(loc=4, prop={'size': 15})


#%% Plot Readout phase
plt.figure(figsize=(8,4))
plt.plot(Phase_ADS_sim[:-1,0],Phase_ADS_sim[:-1,1], '--', linewidth=4, color='k', label='Sim.')
plt.plot(RF_Power[:-1] - 75, Phase_3[0][:-1], '-o', linewidth='3', label='$T_{int}=1024\mu$s',  markersize=8)


plt.ylabel('$\\angle A_{sig}$ [deg.]')
plt.xlabel('$P_{RF}$ [dBm]')
plt.xlim([-125,-80])
plt.ylim([-90, 0])

plt.grid()
plt.legend(loc=0, prop={'size': 15})



# %% SNR Ideal
SNR_1 = 10*np.log10((A_1)/(N_1) / 8e-6) 
SNR_2 = 10*np.log10((A_2)/(N_2) / 128e-6)
SNR_3 = 10*np.log10((A_3)/(N_3) / 1024e-6)

f0 = 6.91e9
T_n_Lim = (h*f0)/(Boltzmann)

T_amb = 0.02 
T_n_lim = (T_n_Lim + T_amb)

N_sim = Boltzmann*T_n_lim*50*G_RX # Quantum limited noise

SNR_sim = 10*np.log10((A_ADS_sim[:,1]**2*G_RX)/N_sim)


plt.figure(figsize=(5,4))
plt.plot(A_ADS_sim[:,0], SNR_sim, '-',linewidth='4', color='k', label='$T_{e}=350$ mK') 

plt.ylabel('$SNR/T_{int}$ [dB.Hz]')
plt.xlabel('$P_{RF}$ [dBm]')
plt.xlim([-125,-65])
plt.ylim([60,100])

# %% SNR Measurement vs Simulation

N_sim = Boltzmann*0.460*50*G_RX # Manually calculated Input referred T_N
N_sim_lim = Boltzmann*T_n_lim*50*G_RX

SNR_sim = 10*np.log10((A_ADS_sim[:,1]**2*G_RX)/N_sim)

plt.figure(figsize=(8,4))
plt.plot(A_ADS_sim[:,0], SNR_sim, '--',linewidth='4', color='k', label='$T_{N}=460$ mK') 

plt.plot(RF_Power[:-1] -75, SNR_1[:-1], '-o', linewidth='2', label='$T_{int}=8\mu$s')
plt.plot(RF_Power[:-1] -75, SNR_2[:-1], '-o', linewidth='2', label='$T_{int}=128\mu$s')
plt.plot(RF_Power[:-1] -75, SNR_3[:-1], '-o', linewidth='2', label='$T_{int}=1024\mu$s')

plt.ylabel('$SNR/T_{int}$ [dB.Hz]')
plt.xlabel('$P_{RF}$ [dBm]')
plt.xlim([-125,-80])
plt.ylim([40,100])

plt.grid()
plt.legend(loc=0, prop={'size': 16})




#%%  2D SNR vs Tint

A_sim_opt = A_ADS_sim[15,1]**2 # [V^2] Readout Signal at -95 dBm 

T_sys_arr = np.linspace(1e-0, 100, num=1000)
T_int_arr = np.linspace(1e-7, 1e-5, num = 1000)
T_amb = 4


SNR_2D_arr = np.empty([len(T_sys_arr),len(T_int_arr)])

for i, T_sys in enumerate(T_sys_arr):
    for j, T_int in enumerate(T_int_arr):
        N = Boltzmann*(T_sys + T_amb)*50 /T_int
        SNR =  10*np.log10(A_sim_opt/N)
        SNR_2D_arr[i,j] = SNR
        
extent = [T_int_arr.min(), T_int_arr.max(), T_sys_arr.min(), T_sys_arr.max()]
cmap = mpl.cm.viridis

X, Y = np.meshgrid(T_int_arr, T_sys_arr)

fig = plt.figure(figsize=(6, 5))

plt.imshow(SNR_2D_arr, origin='lower', extent=extent, aspect = 'auto', cmap=cmap)
plt.xscale('log')

plt.colorbar(label="SNR [dB]", orientation="vertical")
plt.clim(0,30) 


contours = plt.contour(T_int_arr, T_sys_arr, np.transpose(np.transpose(SNR_2D_arr)), levels=np.arange(0,30,4), colors='k', linewidths=2, origin='lower')
plt.yscale('log')

ax = plt.gca()
labels = ax.clabel(contours, inline=True, inline_spacing=10, fontsize=15,  manual=False)

for l in labels:
    l.set_rotation(0)
    

plt.xlabel('$T_{int}$ [s]')
plt.ylabel('$T_{sys}$ [K]')

plt.title('$P_{RF}=-95$dBm, $T_{amb}=4K$')



